from enum import IntEnum
from shapely.geometry import Polygon, LineString, MultiPolygon
from shapely import ops
import numpy as np

import src


class Category(IntEnum):
    UNDEFINED = 0
    LAND = 1
    FIELD = 2
    FOREST = 3
    RIVER = 4
    LAKE = 5
    SEA = 6
    PARK = 7
    GARDEN = 8
    HOUSE = 10
    MANSION = 11
    MARKET = 12
    TOWNHALL = 13
    UNIVERSITY = 14
    FARM = 15
    CHURCH = 20
    CATHEDRAL = 21
    MONASTRY = 22
    FORT = 31
    CASTLE = 32
    WALL = 33
    STREET = 50
    BRIDGE = 51
    COMPOSITE = 90  # a union of Areas


class Area:
    _last_id = 0
    members = []

    @staticmethod
    def get_id():
        Area._last_id += 1
        return Area._last_id

    def __init__(self, polygon, category, sub_areas=[]):
        """
        The Area is the most basic class, however it has all that's needed to
        plot the map.

        Args:
            polygon: Polygon - countour of the area
            category: Category - type of area
            sub_areas: list - list of sub areas if any
        """
        self._polygon = polygon
        self._category = category
        self._sub_areas = sub_areas
        self._id = self.get_id()
        Area.members.append(self)

    def __del__(self):
        try:
            Area.members.remove(self)
        except:
            pass

    def __repr__(self):
        return str(self._category) + ":" + self.polygon.wkt

    @property
    def identity(self):
        return self._id

    @property
    def polygon(self):
        return self._polygon

    def split(self, percentage, direction, inplace=True, new_category=Category.GARDEN):
        """
        Split an area in two areas. Store result in self.sub_areas if inplace == True.

        Args:
            percentage: float - percentage of surface for first area, between 0 and 1
            direction: int - side for first area (from center to 0 = North, 90 = East...)
            new_category: Category - category of the second area

        Returns: if inplace == False
            area1: Area
            area2: Area

        Tests:
            >>> surf = Area(Polygon([(0,0), (20,0), (20,40), (0,40)]), Category.HOUSE)
            >>> res = surf.split(0.25, 0, False)  # house takes 1/4 of surface and is north
            >>> res0 = Polygon([(0, 30), (0, 40), (20, 40), (20, 30), (0, 30)])
            >>> res0.symmetric_difference(res[0].polygon).area < 1
            True
        """
        assert (percentage > 0)
        if not self.polygon.exterior.is_ccw:  # should be counter clockwise
            coords = list(self.polygon.exterior.coords)
            self.polygon = Polygon(coords[::-1])
        direction = np.deg2rad(90 - direction)  # degrees are cw while radian are ccw + 0 is North
        pts = np.array(self.polygon.minimum_rotated_rectangle.exterior.coords)
        diameter = np.sqrt(np.sum(np.square(pts[2] - pts[0])))
        start = np.array(self.polygon.centroid)
        end = start + np.array([diameter * np.cos(direction), diameter * np.sin(direction)])
        path = LineString([start, end])
        pt_intersection = path.intersection(self.polygon.boundary)
        try:
            pt_intersection = list(pt_intersection)[-1]  # we may have more than one intersection
        except:
            pass
        pts = self.polygon.boundary.coords
        for pt1, pt2 in zip(pts[:-1], pts[1:]):
            if LineString((pt1, pt2)).distance(pt_intersection) < 1E-6:
                break
        pt1 = np.array(pt1)
        pt2 = np.array(pt2)
        dist = np.sqrt(np.sum(np.square(pt2 - pt1)))
        dir = (pt2 - pt1) / dist
        orth = np.array([-dir[1], dir[0]])  # ccw
        house_area = self.polygon.area * percentage
        width = diameter / 2  # hence we can reach from 0 to diameter
        res = [self.polygon, ]
        dw = width
        while abs(res[0].area - house_area) > 1:  # 1 meter² error accepted
            cut = LineString([pt1 + width * orth - diameter * dir, pt2 + width * orth + diameter * dir])
            res = ops.split(self.polygon, cut)
            dw /= 2
            if len(res) == 0:
                width -= dw
                continue
            if pt_intersection.distance(res[0]) > pt_intersection.distance(res[1]):
                res = MultiPolygon([res[1], res[0]])
            else:
                res = MultiPolygon([g for g in res])
            if res[0].area > house_area:
                width -= dw
            else:
                width += dw
        area0 = Area(res[0], self._category)
        area1 = Area(res[1], new_category)
        if inplace:
            self._sub_areas = [area0, area1]
        else:
            return area0, area1

    def components(self):
        if len(self._sub_areas) > 0:
            return self._sub_areas
        else:
            return [self, ]


if __name__ == "__main__":
    zone = Area(Polygon([(0, 0), (10, 0), (15, 15), (-5, 10)]), Category.HOUSE)  # units are meters
    zone.split(0.4, 280, inplace=True)  # house in south, it takes 40 % of the area
    src.tools.json(zone, '../tmp/house.json')



__version__ = '0.1.2'

from src.area import Area
from src.castle import Castle
from src.church import Church
from src.city import City
from src.district import District
from src.district import cut_poly
from src.district import voronoi_finite_polygons_2d
from src.district import new_poly
from src.field import Field
from src.forest import Forest
from src.fort import Fort
from src.garden import Garden
from src.house import House
from src.lake import Lake
from src.mansion import Mansion
from src.market import Market
from src.park import Park
from src.street import Street
from src.tools import json
from src.townhall import Townhall
from src.viewer import view
from src.wall import Wall




import src.tools
import numpy as np
import random as rand

from shapely.geometry import Polygon, MultiPolygon
from scipy.spatial import Voronoi

from src.district import District
from src.wall import Wall


class City:
    """
        A class to represent a City.
        This is the main class of this project.
        City is use to create a randomized city with some option that you can write in a Json file.
        ...

        Attributes
        ----------
        population : int
            Population of the city
        density : int
            Density of the city
        districts : list
            List of all district in the city
        has_walls : bool
            If True, the city has walls
        has_castle : bool
            If True, the city has a district castle
        has_lake : bool
            If True, the city has a district lake

        Methods
        -------
        components():
            Get all districts components of the city.
        """

    def __init__(self, population, density=10000, has_walls=False, has_castle=False, has_lake=False, seed=None):
        """
        Constructs all the necessary attributes for the city object.

        Parameters
        ----------
            population : int
                Population of the city
            density : int
                Density of the city
            has_walls : bool
                If True, the city has walls
            has_castle : bool
                If True, the city has a district castle
            has_lake : bool
                If True, the city has a district lake
            seed : int
                Seed for random generation
        """
        np.random.seed(seed)
        rand.seed(seed, 2)

        self.population = population
        self.density = density

        # Some options
        self.has_castle = has_castle
        self.has_lake = has_lake
        self.has_walls = has_walls

        self.districts = []

        # Max number of district per row
        max_district = 8 if density > 2000 else 5
        # Radius of the city
        radius = density / 2
        # Randomized points of districts centers
        points = np.array(
            [[x, y] for x in np.linspace(-1, 1, max_district) for y in np.linspace(-1, 1, max_district)])
        points *= radius
        points += np.random.random((len(points), 2)) * (radius / 6)

        # Make the Voronoi partition of the plane
        vor = Voronoi(points)

        # Get all regions that are in a specific random zone
        zone = Polygon((2 * np.random.random((8, 2)) - 1) * radius).convex_hull.buffer(radius / 2)

        regions = [r for r in vor.regions if -1 not in r and len(r) > 0]
        regions = [Polygon([vor.vertices[i] for i in r]) for r in regions]
        regions = [r for r in regions if zone.contains(r)]

        # Make walls if necessary
        if self.has_walls:
            self.districts.append(Wall(MultiPolygon(regions).buffer(density * 0.008, join_style=1, cap_style=3)))

        # Add all districts in the city
        for polygon in regions:
            if self.has_castle and (rand.randint(0, 100) < 20 or regions[-1] == polygon):
                self.districts.append(District(polygon, 3, density, seed=seed))
                self.has_castle = False
            elif self.has_lake and (rand.randint(0, 100) < 20 or regions[-2] == polygon):
                self.districts.append(District(polygon, 4, density, seed=seed))
                self.has_lake = False
            else:
                r = rand.randint(0, 100)
                self.districts.append(District(polygon, 1 if r > 10 else 2, density, seed=seed))

    def components(self):
        """
        Get the components of the object City

        Parameters
        ----------
        None

        Returns
        -------
        Components of City
        """
        if len(self.districts) > 0:
            return [block.components() for block in self.districts]
        else:
            return []


if __name__ == "__main__":
    city = City(10000, 10000, has_walls=True, has_castle=True, has_lake=True, seed=None)
    # Population is mandatory
    # By default density=10000, has_walls=False, has_castle=False, has_lake=False
    src.tools.json(city, '../tmp/city.json')



from collections import Iterable

from shapely.geometry import mapping
import fiona

SCHEMA = {
    'geometry': 'Polygon',
    'properties': {'category': 'int'},
}


def write(obj, stream):
    if isinstance(obj, Iterable):
        for e in obj:
            write(e, stream)
    else:
        stream.write({
            'geometry': mapping(obj._polygon),
            'properties': {'category': obj._category.value},
        })


def json(what, filename):
    with fiona.open(filename, 'w', 'GeoJSON', SCHEMA) as c:
        write(what.components(), c)



from src.area import *


class Field:
    """
            Field class polygon.

            ...

            Attributes
            ----------
            _polygon : polygon
                polygon of the field
            _area : Area
                Area of the polygon

            Methods
            -------
            components(self):
                Return the component of the Field.
            get_area(self):
                Return the area of the Field.
            get_id():
                Return the id of the area.
            """
    # private:
    _polygon = None
    _area = None

    # public:

    def __init__(self, polygon):
        """
                Constructs all the necessary attributes for the Field object.

                Parameters
                ----------
                    polygon : Polygon
                        Polygon use for the area constructor
                """
        self._polygon = polygon
        self._area = Area(self._polygon, Category.FIELD)

    def components(self):
        """
                Get the components of the object Field

                Parameters
                ----------
                None

                Returns
                -------
                Components of Field
                """
        return self._area.components()

    def get_area(self):
        """
                Get the area of the object Field

                Parameters
                ----------
                None

                Returns
                -------
                Area of Field
                """
        return self._area

    @staticmethod
    def get_id():
        """
                Get the area's id of the object Field

                Parameters
                ----------
                None

                Returns
                -------
                Area's id of Field
                """
        return Area._last_id




from src.area import *


class Castle:
    """
        Castle class polygon.

        ...

        Attributes
        ----------
        _polygon : polygon
            polygon of the castle
        _area : Area
            Area of the polygon

        Methods
        -------
        components(self):
            Return the component of the Castle.
        get_area(self):
            Return the area of the Castle.
        get_id():
            Return the id of the area.
        """
    # private:
    _polygon = None
    _area = None

    # public:

    def __init__(self, polygon):
        """
                Constructs all the necessary attributes for the Castle object.

                Parameters
                ----------
                    polygon : Polygon
                        Polygon use for the area constructor
                """
        self._polygon = polygon
        self._area = Area(self._polygon, Category.CASTLE)

    def components(self):
        """
                Get the components of the object Castle

                Parameters
                ----------
                None

                Returns
                -------
                Components of Castle
                """
        return self._area.components()

    def get_area(self):
        """
                Get the area of the object Castle

                Parameters
                ----------
                None

                Returns
                -------
                Area of Castle
                """
        return self._area

    @staticmethod
    def get_id():
        """
                Get the area's id of the object Castle

                Parameters
                ----------
                None

                Returns
                -------
                Area's id of Castle
                """
        return Area._last_id




from src.area import *


class Street:
    """
                Street class polygon.

                ...

                Attributes
                ----------
                _polygon : polygon
                    polygon of the street
                _area : Area
                    Area of the polygon

                Methods
                -------
                components(self):
                    Return the component of the Street.
                get_area(self):
                    Return the area of the Street.
                get_id():
                    Return the id of the area.
                """
    # private:
    _polygon = None
    _area = None

    # public:

    def __init__(self, polygon):
        """
                Constructs all the necessary attributes for the Street object.

                Parameters
                ----------
                    polygon : Polygon
                        Polygon use for the area constructor
                """
        self._polygon = polygon
        self._area = Area(self._polygon, Category.STREET)

    def components(self):
        """
                Get the components of the object Street

                Parameters
                ----------
                None

                Returns
                -------
                Components of Street
                """
        return self._area.components()

    def get_area(self):
        """
                Get the area of the object Street

                Parameters
                ----------
                None

                Returns
                -------
                Area of Street
                """
        return self._area

    @staticmethod
    def get_id():
        """
                Get the area's id of the object Street

                Parameters
                ----------
                None

                Returns
                -------
                Area's id of Street
                """
        return Area._last_id


import random

from scipy.spatial import Voronoi
import numpy as np
import random as rand

from shapely.geometry import Polygon, Point, MultiPoint

from src.castle import Castle
from src.church import Church
from src.forest import Forest
from src.garden import Garden
from src.house import House
from src.fort import Fort
from src.field import Field
from src.lake import Lake
from src.mansion import Mansion
from src.market import Market
from src.park import Park
from src.street import Street
from src.townhall import Townhall


# List of constructor use in a category 1 district
ctor = [House, Park, Market, Mansion, Townhall, Church]


def cut_poly(polygon, n=50):
    """
    Find random points in a polygon to divide it

    Parameters
    ----------
    polygon : Polygon
        Input polygon
    n : int
        Number of points to find, default is 50

    Returns
    -------
    points : list of tuples
        Coordinates of points in the polygon
    """
    # Polygon bounds
    (min_x, min_y, max_x, max_y) = polygon.bounds
    nb = 0
    points = []
    (min_x, min_y, max_x, max_y) = (round(min_x), round(min_y), round(max_x), round(max_y))

    # Number of pieces the polygon need to be cut
    while nb < n:
        x = rand.randint(min_x, max_x)
        y = rand.randint(min_y, max_y)
        # Check if the polygon contains (x,y) point
        if polygon.contains(Point(x, y)):
            points.append((x, y))
            nb += 1
    return points


def voronoi_finite_polygons_2d(vor, radius=None):
    """
    Reconstruct infinite voronoi regions in a 2D diagram to finite
    regions.

    Parameters
    ----------
    vor : Voronoi
        Input diagram
    radius : float, optional
        Distance to 'points at infinity'.

    Returns
    -------
    regions : list of tuples
        Indices of vertices in each revised Voronoi regions.
    vertices : list of tuples
        Coordinates for revised Voronoi vertices. Same as coordinates
        of input vertices, with 'points at infinity' appended to the
        end.

    """

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge

            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)


def new_poly(polygon, liste, ctor1, ctor2, density, ratio):
    """
        Divide a polygon in two pieces, a border and a center.

        Parameters
        ----------
        polygon : Polygon
            Input Polygon
        liste : list of objects
            List to add the both objects
        ctor1 : constructor
            Constructor of the border object
        ctor2 : constructor
            Constructor of the center object
        density : int
            Density of the city
        ratio : float
            Ratio of border object

        Returns
        -------
        None
    """
    new_pol = polygon.difference(polygon.buffer(-density * ratio, cap_style=2))
    liste.append(ctor1(polygon.buffer(-density * ratio, cap_style=2)))
    liste.append(ctor2(new_pol))


class District:
    """
            A class to represent a City.
            This is the main class of this project.
            City is use to create a randomized city with some option that you can write in a Json file.
            ...

            Attributes
            ----------
            _polygon : Polygon
                Polygon of the district
            _density : int
                Density of the city
            _category : int
                Type of district
            _blocks : List of object
                List of all objects composing the district
            _roads : List of Street
                List of all Streets composing the district

            Methods
            -------
            cut_district(self):
                Cut a category 1 district in multiple object
            components():
                Get the district components.
    """
    # private:
    _polygon = None

    def __init__(self, polygon, category, density, seed=None):
        """
        Constructs all the necessary attributes for the district object.

        Parameters
        ----------
            polygon : Polygon
                Polygon of the district
            density : int
                Density of the city
            category : int
                Category of the district
            seed : int
                Seed for random generation
        """
        np.random.seed(seed)
        rand.seed(seed, 2)

        self._polygon = polygon
        self._category = category
        self._blocks = []
        self._density = density
        self._roads = []
        # Residential district
        if self._category == 1:
            self.cut_district()
        # Field district
        elif self._category == 2:
            self._blocks.append(Field(polygon))
        # Castle district
        elif self._category == 3:
            new_poly(polygon, self._blocks, Castle, Fort, density, 0.02)
        # Lake district
        elif self._category == 4:
            new_poly(polygon, self._blocks, Lake, Forest, density, 0.02)

    def cut_district(self):
        """
            Divide a district in many components pieces.
            From randoms points, make a new voronoi diagram from a definite polygon

            Parameters
            ----------
            None

            Returns
            -------
            None
        """
        # Cut the polygon into pieces with voronoi diagram
        points = cut_poly(self._polygon)
        vor = Voronoi(points)
        x, y = self._polygon.exterior.xy
        points = [(x[i], y[i]) for i in range(len(x))]

        # Get regions from the voronoi diagram
        regions, vertices = voronoi_finite_polygons_2d(vor)

        # Make a mask to cut regions at the border district
        pts = MultiPoint([Point(i) for i in points])
        mask = pts.convex_hull

        for region in regions:
            polygon = vertices[region]
            shape = list(polygon.shape)
            shape[0] += 1
            # Get the intersection between the region and the polygon
            p = Polygon(np.append(polygon, polygon[0]).reshape(*shape)).intersection(mask)
            new_vertices = np.array(list(zip(p.boundary.coords.xy[0][:-1], p.boundary.coords.xy[1][:-1])))
            tmp = [(x, y) for x, y in new_vertices]
            # Create a new polygon
            poly = Polygon(tmp)
            # Create Street from it
            road = poly.difference(poly.buffer(-self._density * 0.002, join_style=1, cap_style=3))
            index = 0 if random.randint(0, 100) > 3 else random.randint(1, len(ctor) - 1)
            # Generate a House 97% of the time
            # Else create a random building
            if index == 0:
                new_poly(poly, self._blocks, Garden, House, self._density, 0.009)
            else:
                self._blocks.append(ctor[index](poly.buffer(-self._density * 0.002, join_style=2, cap_style=3)))
            self._blocks.append(Street(road))

    def components(self):
        """
        Get the components of the object District

        Parameters
        ----------
        None

        Returns
        -------
        Components of District
        """
        if len(self._blocks) > 0:
            return [block.components() for block in self._blocks]
        else:
            return []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['src', 'src.pton_pgk', 'tests']

package_data = \
{'': ['*'], 'tests': ['.ipynb_checkpoints/*']}

install_requires = \
['Shapely>=1.7.1,<2.0.0', 'geopandas>=0.9.0,<0.10.0', 'numpy>=1.21.0,<2.0.0']

setup_kwargs = {
    'name': 'pton-pgk',
    'version': '0.1.2',
    'description': 'Medieval City',
    'long_description': "# PTON Project : City Generator\n**Marius PON** et **Tom THIL**, *EPITA Promo 2023*\n\nThis projects consists in a medieval city generator in python. The generated city depends on attributes as the number of population and its density.\n\n## Features\n- Generate a medieval city depending on the number of population and the density of the population\n- It is possible to generate or not a castle, a wall and a lake in the city\n- The result is stored in a *.json* file\n- The user has to possibility to use a city viewer\n\n\n## How to use\n\n### Instantiate a 'City' object\nThis project can be executed with your *Pycharm* IDE. You have to run the *city.py* file, located in the *src/* repository.\n\nA city object is instanciated as follows :\n``city = City(population, density, has_walls, has_castle, has_lake)``\n\nA City object has already been created in *city.py* with values giving good results.  You can custom the generated cities by replacing them by your own values.\n\nThere is a last *'seed'* parameter, none as default. 2 cities generated with the same parameters and the same seeds will be identical.\n\n### Generating a *.json* file\nYou can generate a *.json* file for your 'city' object as follows :\n``tools.json(city, 'path/to/your/file/.json')``\n\n### Use the city viewer\nTo use the **city viewer**, you have to run run *viewer.py*. There is a *view* function taking as parameter a *.json* file corresponding to a generated city. By default, the fonction takes *tmp/city.json* as argument.\n\n## Documentation\nA documentation has already been generated (in *docs/_build/html/tml/index.html*).\n\nHowever, if you want to generate it again for some reason, you must run the command ``make html`` in the *docs* folder.",
    'author': 'Tom',
    'author_email': 'tom.thil@epita.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
